%steganography
% Function which hides a text file within a PNG image, or, alternatively,
% extracts a hidden message from a PNG image. The idea is that the ASCII
% codes for text characters are broken up into digits and added to the
% 'strength' of R,G,B colour planes.
%
% LAST UPDATED by Andy French April 2016

function steganography

%Display message to command window
disp(' ')
disp(' %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
disp('  Steganography - how to hide messages in digital photos')
disp(' %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
disp(' ');

% Ask user for the original .PNG file
disp(' ... Select original PNG image')
disp(' ');
[orig_image_filename,orig_image_pathname] = uigetfile( '*.png',pwd );
if orig_image_filename==0
    return
end

%Load image file
I = imread( [orig_image_pathname,orig_image_filename] );

%Step through colour planes and make sure maximum intensity of R,G,B is
%255-9 = 245. 9 is the maximum change the message addition will produce.
I( I > 245 ) = 245;

%Extract colour planes
[R,G,B] = extract_RGB(I);

% Choose Encrypt or Decrypt mode
mode_choice = questdlg('Choose Encrypt or Decrypt modes',...
    'Steganography','Encrypt','Decrypt','Decrypt');

switch mode_choice
    case 'Encrypt'
        
        %% Encrypt mode %%
        
        %Display message to command window
        disp(' ENCRYPTING ...')
        
        %Ask user for .txt file which contains the message to encrypt
        [text_filename,text_pathname] = uigetfile( '*.txt',pwd );
        
        %Start clock
        t0 = tic;
        
        %Load text file into a row vector of ASCII characters
        fid = fopen( [text_pathname,text_filename,], 'r' );
        text_chars = fscanf(fid, '%c');
        fclose(fid); clear fid;
        
        %Convert characters into ASCII codes (i.e. integers)
        text_ascii = double(text_chars); 
        clear text_chars;
        
        %Find number of characters in message
        N = numel(text_ascii);
        
        %Truncate message if bigger than number of pixels in image
        if N > numel(R)
            N = numel(R);
            text_ascii = text_ascii(1:N);
            warndlg('Image not large enough to store message, truncated to fit',...
                'Steganographic warning!');
        end
        
        %Step through text_ascii and assign units, tens and hundreds
        %integer of the ASCII code (assume no codes above 999) to,
        %respective, R,G,B colour planes of orig_image
        y = idigits( text_ascii, 3 );
        R(1:N) = R(1:N) + y{1};
        G(1:N) = G(1:N) + y{2};
        B(1:N) = B(1:N) + y{3};
        
        %Export image - resuse I for original image
        I(:,:,1) = R(:,:);
        I(:,:,2) = G(:,:);
        I(:,:,3) = B(:,:);
        imwrite( I, [orig_image_pathname,'encrypted ',orig_image_filename],'png' );
        clear R G B
        
        %Stop clock and display completion message
        disp( [' Message ',text_filename,' encrpted into image ',orig_image_filename,...
            ' in ',num2str(toc(t0)),' seconds.' ] );
        disp(' ');
        
    case 'Decrypt'
        
        %% Decrypt mode %%
              
        %Choose encrypted image file
        [encrypt_image_filename,encrypt_image_pathname] = uigetfile( '*.png',pwd );   
        
        %Start clock and load encrypted image file
        t0 = tic;
        encrypt_image = imread( [encrypt_image_pathname,encrypt_image_filename] );
        
        %Display message to command window
        disp(' DECRYPTING ...')
        
        %Extract colour planes
        [eR,eG,eB] = extract_RGB( encrypt_image );
        
        %Determine difference between encrypted and original colour planes
        dR = eR - R;
        dG = eG - G;
        dB = eB - B;
        
        %Create vector as ascii codes
        text_ascii = 100*dR + 10*dG + dB;
        clear R G B dR dG dB eR eG eB
        
        %Remove zeros
        text_ascii( text_ascii == 0 ) = [];
        
        %Convert text_ascii to a text string
        text_chars = char( text_ascii );
        
        %Write out message file
        fid = fopen( [encrypt_image_pathname,'decrypt ',encrypt_image_filename(1:end-3),'txt'], 'w' );
        fwrite( fid, text_chars );
        fclose(fid); clear fid;
        
        %Stop clock and display completion message
        disp( [' Message extracted from image ',encrypt_image_filename,...
            ' in ',num2str(toc(t0)),' seconds.' ] );
        disp(' ');
end

%%

%extract_RGB
% Extract R,G,B uint8 matrices from a uint8 image matrix I

function [R,G,B] = extract_RGB(I)
I = uint8(I);
dim = size(I);
R = uint8( zeros(dim(1),dim(2)) );
G = uint8( zeros(dim(1),dim(2)) );
B = uint8( zeros(dim(1),dim(2)) );
if dim(3)==1
    %Black & white image
    R(:,:) = I(:,:,1);
    G = R;
    B = R;
else
    %Full colour image
    R(:,:) = I(:,:,1);
    G(:,:) = I(:,:,2);
    B(:,:) = I(:,:,3);
end

%%

%idigits
% Returns a cell array containing (uint8) integer values of digits in a decimal
% representation. e.g. idigits( 123, 4 ) = { [0],[1],[2],[3] };

function y = idigits(x,N)

%Define output
y = cell(1,N);

%Step through digits and populate output
A = zeros(size(x));
B = x;
for n=1:N
    %Remove previous digits
    B = B - A;
    
    %Work out this digit and store in output
    A = mod( B, 10^n );
    y{ N - n + 1} = uint8( A/(10^(n-1)) );
end

% End of code